@@ -28,6 +28,7 @@
 #include <string.h>
 #include <ctype.h>
 #include <sched.h>
+#include <sys/statvfs.h>
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -51,9 +52,10 @@
 #include <nih-dbus/dbus_proxy.h>
 
 #include "frontend.h"  // for keys_return_type
+#include "fs.h"        // for #defines
 
 /* defines relating to the release agent */
-#define AGENT SBINDIR "/cgm-release-agent"
+#define AGENT LIBEXECDIR "/cgmanager/cgm-release-agent"
 #define AGENT_LINK_PATH "/run/cgmanager/agents"
 
 /* Define pivot_root() if missing from the C library */
@@ -83,6 +85,7 @@ struct controller_mounts {
 	bool premounted;
 	bool visited;
 	bool skip;
+	bool unified;
 };
 
 static struct controller_mounts *all_mounts;
@@ -140,10 +143,6 @@ bool setup_base_run_path(void)
 		nih_fatal("%s: failed to create /run/cgmanager/fs", __func__);
 		return false;
 	}
-	if (mount("cgmfs", "/run/cgmanager/fs", "tmpfs", 0, "size=100000,mode=0755") < 0) {
-		nih_fatal("%s: failed to mount tmpfs onto /run/cgmanager/fs", __func__);
-		return false;
-	}
 	if (mkdir(AGENT_LINK_PATH, 0755) < 0 && errno != EEXIST) {
 		nih_fatal("%s: failed to create %s", __func__, AGENT_LINK_PATH);
 		return false;
@@ -246,19 +245,10 @@ static bool fill_in_controller(struct controller_mounts *m, char *controller,
 	nih_local char *dest = NULL;
 
 	dest = NIH_MUST( nih_sprintf(NULL, "%s/%s", base_path, src) );
-	m->controller = strdup(controller);
-	if (!m->controller) {
-		nih_fatal("Out of memory mounting controllers");
-		return false;
-	}
+	m->controller = NIH_MUST( strdup(controller) );
 	m->options = NULL;
-	m->path = strdup(dest);
-	m->src = strdup(src);
-	if (!m->path ||
-			!m->src) {
-		nih_fatal("Out of memory mounting controllers");
-		return false;
-	}
+	m->path = NIH_MUST( strdup(dest) );
+	m->src = NIH_MUST( strdup(src) );
 	nih_info(_("Arranged to mount %s onto %s"), m->controller, m->path);
 	return true;
 }
@@ -295,7 +285,6 @@ static bool save_mount_subsys(char *s)
 	insert_pt = find_controller_in_mounts(controller, &found);
 	if (found)
 		return true;
-
 	tmp = realloc(all_mounts, (num_controllers+1) * sizeof(*all_mounts));
 	if (!tmp) {
 		nih_fatal("Out of memory mounting controllers");
@@ -366,6 +355,16 @@ static bool do_mount_subsys(int i)
 		nih_fatal("Failed to create %s: %s", dest, strerror(errno));
 		return false;
 	}
+
+	if (m->unified) {
+		if (mount(m->controller, dest, "cgroup", 0, "__DEVEL__sane_behavior") < 0) {
+			nih_error("Failed mounting %s: %s\n", m->controller,
+					strerror(errno));
+			return false;
+		}
+		return true;
+	}
+
 	if (m->premounted)
 		ret = mount(src, dest, "cgroup", 0, m->options);
 	else
@@ -426,8 +425,11 @@ static bool process_mounted_subsystem(char *options)
 	while (tok) {
 		if (strncmp(tok, "name=", 5) == 0) {
 			i = find_controller_in_mounts(tok+5, &found);
-			if (found) // jinkeys, multiple mounts already
+			if (found) {
+				if (all_mounts[i].unified)
+					return true;
 				goto next;
+			}
 			if (!save_mount_subsys(tok))
 				return false;
 			i = find_controller_in_mounts(tok+5, &found);
@@ -440,8 +442,11 @@ static bool process_mounted_subsystem(char *options)
 				NIH_MUST( nih_strcat_sprintf(&cp_opts, NULL, ",%s", tok) );
 		} else if (is_kernel_controller(tok)) {
 			i = find_controller_in_mounts(tok, &found);
-			if (found) // jinkeys, multiple mounts already
+			if (found) {
+				if (all_mounts[i].unified)
+					return true;
 				goto next;
+			}
 			if (!save_mount_subsys(tok))
 				return false;
 			i = find_controller_in_mounts(tok, &found);
@@ -532,6 +537,8 @@ static bool collate_premounted_subsystems(void)
 		first = &all_mounts[i];
 		if (!first->premounted)
 			continue;
+		if (first->unified)
+			continue;
 		if (first->comounted) // already linked
 			continue;
 		if (!first->options)
@@ -626,8 +633,12 @@ again:
 	/* now we know for sure that [f-1,f+len+1] == ",f," */
 	if (f[len])
 		memmove(f, f+len+1, strlen(f+len+1)+1);
-	else
-		*(f-1) = '\0';
+	else {
+		if (f > origlist)
+			*(f-1) = '\0';
+		else
+			*f = '\0';
+	}
 	goto again;
 }
 
@@ -696,6 +707,7 @@ static void print_debug_controller_info(void)
 		nih_debug("    premounted: %s comounted: %s",
 			m->premounted ? "yes" : "no",
 			m->comounted ? m->comounted->controller : "(none)");
+		nih_debug("    unified: %s", m->unified ? "yes" : "no");
 	}
 }
 
@@ -706,7 +718,7 @@ void do_list_controllers(void *parent, char ***output)
 	nih_assert(output);
 	*output = NIH_MUST( nih_alloc(parent, (num_controller_mnts+1) * sizeof(char *)) );
 	(*output)[num_controller_mnts] = NULL;
-	
+
 	/* XXX
 	 * This will actually not be right.
 	 * if we have freezer,devices co-mounted, we'll have two separate
@@ -745,6 +757,26 @@ void do_prune_comounts(char *controllers)
 	}
 }
 
+/*
+ * If we are passed 'cpu', do nothing and return true.
+ * If we are passed 'cpu,cpuacct,devices', and all three are comounted,
+ *   set controllers to 'cpu' and return true.
+ * If we are passed 'cpu,cpuacct,devices', and all three are not comounted,
+ *   return false.
+ */
+bool prune_verify_comounts(char *controllers)
+{
+	char *comma;
+	do_prune_comounts(controllers);
+	comma = strchr(controllers, ',');
+	if (!comma)
+		return true;
+	if (*(comma+1) != '\0')
+		return false;
+	*comma = '\0';
+	return true;
+}
+
 /*
  * @list is a comma-separated list of words.
  * Return true if @word is in @list.
@@ -794,9 +826,257 @@ static void build_all_controllers(char *skip_mounts)
 	do_prune_comounts(all_controllers);
 }
 
+/*
+ * Check whether the unified hierarchy is available
+ */
+static bool unified_hierarchy_present(void)
+{
+	FILE *f;
+	char *line = NULL;
+	size_t len = 0;
+	bool ret = false;
+
+	if ((f = fopen("/proc/self/cgroup", "r")) == NULL)
+		return false;
+
+	while (getline(&line, &len, f) != -1) {
+		if (strncmp(line, "0:", 2) == 0) {
+			ret = true;
+			break;
+		}
+	}
+
+	fclose(f);
+	free(line);
+	return ret;
+}
+
+/*
+ * Mount a transient instance of the unified hierarchy, so that
+ * we can pin the controllers currently enabled there
+ */
+static bool mount_transient_unified(void)
+{
+	if (mkdir(UNIFIED_DIR, 0755) < 0 && errno != EEXIST)
+		return false;
+	if (mount("cgroup", UNIFIED_DIR, "cgroup", 0, "__DEVEL__sane_behavior") < 0) {
+		nih_error("Error mounting unified: %s\n", strerror(errno));
+		return false;
+	}
+	return true;
+}
+
+static void mark_unified_controllers_comounted(void)
+{
+	int i;
+	struct controller_mounts *first = NULL, *last = NULL;
+
+	for (i = 0; i < num_controllers; i++) {
+		if (!all_mounts[i].unified)
+			continue;
+		if (!first) {
+			first = last = &all_mounts[i];
+			continue;
+		}
+		last->comounted = &all_mounts[i];
+		last = &all_mounts[i];
+	}
+	if (last && last != first)
+		last->comounted = first;
+}
+
+static bool record_unified_controllers(char *ctrl_list)
+{
+	struct controller_mounts *tmp;
+	int i, insert_pt;
+	bool found;
+	char *tok;
+
+	tok = strtok(ctrl_list, " ");
+	while (tok) {
+		insert_pt = find_controller_in_mounts(tok, &found);
+		if (found) {
+			/*
+			 * Something in the program flow is not right.
+			 * We must not know what's actually going on.
+			 */
+			nih_error("Impossible: found duplicate in unified (%s)", tok);
+			return false;
+		}
+
+		tmp = realloc(all_mounts, (num_controllers+1) * sizeof(*all_mounts));
+		if (!tmp) {
+			nih_fatal("Out of memory mounting controllers");
+			return false;
+		}
+		all_mounts = tmp;
+
+		for (i = num_controllers; i > insert_pt; i--)
+			all_mounts[i] = all_mounts[i-1];
+		zero_out(&all_mounts[insert_pt]);
+
+		if (!fill_in_controller(&all_mounts[insert_pt], tok, tok))
+			return false;
+		all_mounts[insert_pt].unified = true;
+		num_controllers++;
+		tok = strtok(NULL, " ");
+	}
+
+	return true;
+}
+
+static inline void drop_newlines(char *s)
+{
+	int l;
+
+	for (l=strlen(s); l>0 && s[l-1] == '\n'; l--)
+		s[l-1] = '\0';
+}
+
+static char *read_oneline(const char *from)
+{
+	char *line = NULL;
+	FILE *f = fopen(from, "r");
+	size_t len = 0;
+	if (!f)
+		return NULL;
+	if (getline(&line, &len, f) == -1) {
+		fclose(f);
+		return NULL;
+	}
+	fclose(f);
+
+	drop_newlines(line);
+	return line;
+}
+
+static bool write_string(const char *fnam, char *string)
+{
+	FILE *f;
+	size_t len, ret;
+
+	if (!(f = fopen(fnam, "w")))
+		return false;
+	len = strlen(string);
+	ret = fwrite(string, 1, len, f);
+	if (ret != len) {
+		nih_error("Error writing to file: %s", strerror(errno));
+		fclose(f);
+		return false;
+	}
+	if (fclose(f) < 0) {
+		nih_error("Error writing to file: %s", strerror(errno));
+		return false;
+	}
+	return true;
+}
+
+static bool pin_and_process_unified(void)
+{
+	nih_local char  *ctrlcopy = NULL,
+			*ctrlline = NULL;
+	char path[MAXPATHLEN];
+	char *line = NULL, *tok;
+	FILE *f;
+
+	if (mkdir(UNIFIED_PIN, 0755) < 0 && errno != EEXIST)
+		return false;
+
+	sprintf(path, "%s/cgroup.controllers", UNIFIED_DIR);
+	line = read_oneline(path);
+	if (!line || strlen(line) == 0) {
+		free(line);
+		return true;
+	}
+
+	ctrlcopy = NIH_MUST( nih_strdup(NULL, line) );
+	tok = strtok(line, " ");
+	while (tok) {
+		NIH_MUST( nih_strcat_sprintf(&ctrlline, NULL, "+%s ", tok) );
+		tok = strtok(NULL, " ");
+	}
+	free(line);
+
+	sprintf(path, "%s/cgroup.subtree_control", UNIFIED_DIR);
+	if (!write_string(path, ctrlline)) {
+		nih_error("Error pinning unified controllers");
+		return false;
+	}
+
+	sprintf(path, "%s/cgroup.subtree_control", UNIFIED_PIN);
+	if (!write_string(path, ctrlline)) {
+		nih_error("Error pinning unified controllers");
+		return false;
+	}
+
+	sprintf(path, "%s/cgroup.procs", UNIFIED_DIR);
+	if (!(f = fopen(path, "w")))
+		return true;
+	fprintf(f, "%d", (int)getpid());
+	fclose(f);
+
+	nih_debug("pinned the following controllers unified: %s\n", ctrlline);
+
+	return record_unified_controllers(ctrlcopy);
+}
+
+static bool do_mount_unified(void)
+{
+	int i;
+	bool found = false;
+	nih_local char *dest = NULL;
+
+	for (i = 0; i < num_controllers; i++) {
+		if (all_mounts[i].unified) {
+			found = true;
+			break;
+		}
+	}
+	if (!found)
+		return true;
+
+	dest = NIH_MUST( nih_sprintf(NULL, "%s/%s", base_path, ".cgm_unified") );
+	if (mkdir(dest, 0755) < 0 && errno != EEXIST) {
+		nih_fatal("Failed to create %s: %s", dest, strerror(errno));
+		return false;
+	}
+
+	if (mount("unified", dest, "cgroup", 0, "__DEVEL__sane_behavior") < 0) {
+		nih_fatal("Failed to mount unified hierarchy: %s\n", strerror(errno));
+		return false;
+	}
+
+	return true;
+}
+
+static void umount_transient_unified(void)
+{
+	umount(UNIFIED_DIR);
+	rmdir(UNIFIED_DIR);
+}
+
+static bool process_unified_hierarchy(void)
+{
+	bool ret = false;
+
+	if (!unified_hierarchy_present())
+		return true;
+	if (!mount_transient_unified())
+		return false;
+	if (pin_and_process_unified())
+		ret = true;
+
+	umount_transient_unified();
+	return ret;
+}
+
 int collect_subsystems(char *extra_mounts, char *skip_mounts)
 {
-	/* first collect all already-mounted subsystems */
+	/* first mount and pin anything currently in the unified hierarchy */
+	if (!process_unified_hierarchy())
+		return -1;
+
+	/* next collect all already-mounted subsystems */
 	if (!collect_premounted_subsystems())
 		return -1;
 
@@ -817,6 +1097,7 @@ int collect_subsystems(char *extra_mounts, char *skip_mounts)
 	if (!collate_premounted_subsystems())
 		return -1;
 
+	mark_unified_controllers_comounted();
 	build_all_controllers(skip_mounts);
 
 	build_controller_mntlist();
@@ -888,41 +1169,41 @@ fail:
 
 static int pivot_into_new_root(void) {
 	int i, ret;
-	char *createdirs[] = {NEWROOT "/proc", NEWROOT "/run",
-		NEWROOT "/run/cgmanager", NEWROOT "/run/cgmanager/fs", NULL};
-	char path[100];
+	char *createdirs[] = {NEWROOT "/run/cgmanager", NEWROOT "/run/cgmanager/fs", NULL};
 
 	/* Mount tmpfs for new root */
 	if (mkdir(NEWROOT, 0755) < 0 && errno != EEXIST) {
 		nih_fatal("%s: Failed to create directory for new root\n", __func__);
 		return -1;
 	}
-	ret = mount("root", NEWROOT, "tmpfs", 0, "size=10000,mode=0755");
+	ret = mount("/", NEWROOT, NULL, MS_BIND, 0);
 	if (ret < 0) {
-		nih_fatal("%s: Failed to mount tmpfs for root", __func__);
+		nih_fatal("%s: Failed to bind-mount / for new root", __func__);
 		return -1;
 	}
 
-	/* create /proc and /run/cgmanager/fs, and move-mount those */
-	for (i = 0; createdirs[i]; i++) {
-		if (mkdir(createdirs[i], 0755) < 0) {
-			nih_fatal("%s: failed to created %s\n", __func__, createdirs[i]);
-			return -1;
-		}
-	}
-
-	ret = snprintf(path, 100, NEWROOT "/proc");
-	if (ret < 0 || ret > 100)
-		return -1;
-	if (mount("/proc", path, NULL, MS_REC|MS_MOVE, 0) < 0) {
+	if (mount("/proc", NEWROOT "/proc", NULL, MS_REC|MS_MOVE, 0) < 0) {
 		nih_fatal("%s: failed to move /proc into new root: %s",
 			__func__, strerror(errno));
 		return -1;
 	}
-	ret = snprintf(path, 100, NEWROOT "/run/cgmanager/fs");
-	if (ret < 0 || ret > 100)
+
+	ret = mount("/run", NEWROOT "/run", NULL, MS_BIND, 0);
+	if (ret < 0) {
+		nih_fatal("%s: failed to move /run into new root: %s",
+			__func__, strerror(errno));
 		return -1;
-	if (mount("/run/cgmanager/fs", path, NULL, MS_REC|MS_MOVE, 0) < 0) {
+	}
+
+	for (i = 0; createdirs[i]; i++) {
+		if (mkdir(createdirs[i], 0755) < 0 && errno != EEXIST) {
+			nih_fatal("%s: failed to create %s : %s\n", __func__,
+				createdirs[i], strerror(errno));
+			return -1;
+		}
+	}
+
+	if (mount("/run/cgmanager/fs", NEWROOT "/run/cgmanager/fs", NULL, MS_REC|MS_MOVE, 0) < 0) {
 		nih_fatal("%s: failed to move /run/cgmanager/fs into new root: %s",
 			__func__, strerror(errno));
 		return -1;
@@ -961,16 +1242,14 @@ int setup_cgroup_mounts(void)
 		return -1;
 	}
 
-	/*
-	 * Mount a tmpfs on top of /root in case / is still ro when we
-	 * started, so we can write the org_freedesktop_general.lock
-	 */
-	if (dir_exists("/root")) {
-		if (mount("root", "/root", "tmpfs", 0, "size=10000") < 0) {
-			nih_warn("Failed to mount a writeable tmpfs onto /root");
-		}
+	if (mount("cgmfs", "/run/cgmanager/fs", "tmpfs", 0, "size=100000,mode=0755") < 0) {
+		nih_fatal("%s: failed to mount tmpfs onto /run/cgmanager/fs", __func__);
+		return false;
 	}
 
+	if (!do_mount_unified())
+		return -1;
+
 	for (i=0; i<num_controllers; i++) {
 		if (!do_mount_subsys(i)) {
 			nih_fatal("Failed mounting cgroups");
@@ -984,6 +1263,16 @@ int setup_cgroup_mounts(void)
 		return -1;
 	}
 
+	/*
+	 * Mount a tmpfs on top of /root in case / is still ro when we
+	 * started, so we can write the org_freedesktop_general.lock
+	 */
+	if (dir_exists("/root")) {
+		if (mount("root", "/root", "tmpfs", 0, "size=10000") < 0) {
+			nih_warn("Failed to mount a writeable tmpfs onto /root");
+		}
+	}
+
 	return 0;
 }
 
@@ -1024,6 +1313,7 @@ bool create_agent_symlinks(void)
 		}
 		nih_info("buf is %s", buf);
 		if (!file_exists(buf)) {
+			unlink(buf); // in case link dest moved
 			if (symlink(AGENT, buf) < 0) {
 				nih_error("failed to create release agent for %s",
 					all_mounts[i].controller);
@@ -1039,14 +1329,6 @@ bool create_agent_symlinks(void)
 	return true;
 }
 
-static inline void drop_newlines(char *s)
-{
-	int l;
-
-	for (l=strlen(s); l>0 && s[l-1] == '\n'; l--)
-		s[l-1] = '\0';
-}
-
 /*
  * The user will pass in 'cpuset' or 'systemd'.  /proc/self/cgroup will
  * show 'cpuset:' or 'name=systemd:'.  We have to account for that.
@@ -1062,6 +1344,26 @@ static bool is_same_controller(const char *cmp, const char *cnt)
 	return false;
 }
 
+/*
+ * In unified hierarchy tasks must be in a leaf node.  Cgmanager
+ * creates .cgm_leaf for tasks.  Ignore that.
+ */
+static void chop_leaf(char *path)
+{
+	size_t len;
+	char *cmp;
+
+	if (!path)
+		return;
+
+	len = strlen(path);
+	if (len < strlen(U_LEAF))
+		return;
+	cmp = path + len - strlen(U_LEAF_NAME);
+	if (strcmp(cmp, U_LEAF_NAME) == 0)
+		*cmp = '\0';
+}
+
 /*
  * pid_cgroup: return the cgroup of @pid for @controller.
  * retv must be a (at least) MAXPATHLEN size buffer into
@@ -1103,6 +1405,8 @@ static inline char *pid_cgroup(pid_t pid, const char *controller, char *retv)
 found:
 	fclose(f);
 	free(line);
+	if (is_unified_controller(controller))
+		chop_leaf(cgroup);
 	return cgroup;
 }
 
@@ -1192,7 +1496,7 @@ bool may_access(pid_t pid, uid_t uid, gid_t gid, const char *path, int mode)
 
 	ret = stat(path, &sb);
 	if (ret < 0) {
-		nih_error("Could not look up %s\n", path);
+		nih_debug("Could not look up %s\n", path);
 		return false;
 	}
 
@@ -1253,6 +1557,16 @@ const char *get_controller_path(const char *controller)
 	return all_mounts[i].path;
 }
 
+bool is_unified_controller(const char *controller)
+{
+	int i;
+	for (i = 0; i < num_controllers; i++)
+		if (strcmp(all_mounts[i].controller, controller) == 0 &&
+				all_mounts[i].unified)
+			return true;
+	return false;
+}
+
 int get_path_depth(const char *p)
 {
 	int depth = 0;
@@ -1340,6 +1654,70 @@ bool compute_pid_cgroup(pid_t pid, const char *controller, const char *cgroup,
 	return true;
 }
 
+#define SYSTEMD_INIT_SLICE "/init.scope"
+#define SYSTEMD_CGPROXY_SLICE "/system.slice/cgproxy.service"
+
+static inline bool check_init_slice(char *path, size_t pathlen)
+{
+	size_t slen = strlen(SYSTEMD_INIT_SLICE);
+	char *point;
+
+	if (pathlen <= slen)
+		return false;
+
+	point = path + pathlen - slen;
+	if (strncmp(point, SYSTEMD_INIT_SLICE, slen) == 0) {
+		if (path == point) // keep a single '/'
+			*(point+1) = '\0';
+		else // drop trailling '/'
+			*point = '\0';
+		return true;
+	}
+	return false;
+}
+
+static inline bool check_cgproxy_slice(char *path, size_t pathlen)
+{
+	size_t slen = strlen(SYSTEMD_CGPROXY_SLICE);
+	char *point;
+
+	if (pathlen <= slen)
+		return false;
+
+	point = path + pathlen - slen;
+	if (strncmp(point, SYSTEMD_CGPROXY_SLICE, slen) == 0) {
+		if (path == point) // keep a single '/'
+			*(point+1) = '\0';
+		else // drop trailling '/'
+			*point = '\0';
+		return true;
+	}
+	return false;
+}
+
+/*
+ * compute_proxy_cgroup - same as compute_pid_cgroup, but chops of the
+ * final /system.slice/cgproxy.service.
+ */
+bool compute_proxy_cgroup(pid_t pid, const char *controller, const char *cgroup,
+		char *path, int *depth)
+{
+	size_t pathlen;
+
+	if (!compute_pid_cgroup(pid, controller, cgroup, path, depth))
+		return false;
+
+	pathlen = strlen(path);
+
+	if (check_init_slice(path, pathlen))
+		return true;
+
+	if (check_cgproxy_slice(path, pathlen))
+		return true;
+
+	return true;
+}
+
 /*
  * file_read_string:
  *
@@ -1585,12 +1963,12 @@ bool chown_cgroup_path(const char *path, uid_t uid, gid_t gid, bool all_children
 		nih_local char *fpath = NULL;
 		fpath = NIH_MUST( nih_sprintf(NULL, "%s/cgroup.procs", path) );
 		if (chown(fpath, uid, gid) < 0)
-			nih_error("Failed to chown procs file %s: %s", fpath,
-				strerror(errno));
+			nih_error("%s: Failed to chown procs file %s: %s", __func__,
+					fpath, strerror(errno));
 		sprintf(fpath+len, "/tasks");
 		if (chown(fpath, uid, gid) < 0)
-			nih_error("Failed to chown tasks file %s: %s", fpath,
-				strerror(errno));
+			nih_warn("%s: Failed to chown the tasks file %s: %s\n",
+					__func__, fpath, strerror(errno));
 	}
 
 out:
@@ -1656,9 +2034,11 @@ bool set_value_trusted(const char *path, const char *value)
 	}
 	return true;
 }
-bool set_value(const char *path, const char *value)
+bool set_value(const char *controller, const char *path, const char *value)
 {
 	int i;
+	char *p;
+	nih_local char *upath = NULL, *file = NULL;
 
 	nih_assert (path);
 
@@ -1674,7 +2054,24 @@ bool set_value(const char *path, const char *value)
 		}
 	}
 
-	return set_value_trusted(path, value);
+	if (!set_value_trusted(path, value))
+		return false;
+
+	if (!is_unified_controller(controller))
+		return true;
+	upath = NIH_MUST( nih_strdup(NULL, path) );
+	p = strrchr(upath, '/');
+	if (!p)
+		return false;  // can't happen
+	file = NIH_MUST( nih_strdup(NULL, p+1) );
+	*p = '\0';
+	NIH_MUST( nih_strcat(&upath, NULL, U_LEAF) );
+	if (!dir_exists(upath))
+		return true;
+	NIH_MUST( nih_strcat(&upath, NULL, "/") );
+	NIH_MUST( nih_strcat(&upath, NULL, file) );
+
+	return (set_value_trusted(upath, value));
 }
 
 /*
@@ -1748,6 +2145,8 @@ bool move_self_to_root(void)
 			continue;
 		if (all_mounts[i].skip)
 			continue;
+		if (all_mounts[i].unified)
+			continue;
 		path = NIH_MUST( nih_sprintf(NULL, "%s/tasks", all_mounts[i].path) );
 		if ((f = fopen(path, "w")) == NULL)
 			return false;
@@ -1793,6 +2192,8 @@ int get_directory_children(void *parent, const char *path, char ***output)
 	while (readdir_r(d, &dirent, &direntp) == 0 && direntp) {
 		if (!strcmp(direntp->d_name, ".") || !strcmp(direntp->d_name, ".."))
 			continue;
+		if (!strcmp(direntp->d_name, U_LEAF_NAME))
+			continue;
 		if (direntp->d_type != DT_DIR)
 			continue;
 		if (used+1 >= alloced) {
@@ -1903,13 +2304,13 @@ bool was_premounted(const char *controller)
 /*
  * Check that (absolute) @path is under @pid's cgroup for @contr
  */
-bool path_is_under_taskcg(pid_t pid, const char *contr,const char *path)
+bool path_is_under_proxycg(pid_t pid, const char *contr,const char *path)
 {
 	char pcgpath[MAXPATHLEN];
 	size_t plen;
 
 	// Get p's current cgroup in pcgpath
-	if (!compute_pid_cgroup(pid, contr, "", pcgpath, NULL)) {
+	if (!compute_proxy_cgroup(pid, contr, "", pcgpath, NULL)) {
 		nih_error("%s: Could not determine the proxy's cgroup for %s",
 				__func__, contr);
 		return false;
@@ -1930,3 +2331,140 @@ bool path_is_under_taskcg(pid_t pid, const char *contr,const char *path)
 		return true;
 	return false;
 }
+
+static bool do_copy_controllers(const char *from, const char *to)
+{
+	nih_local char  *src = NULL,
+		  *dest = NULL,
+		  *ctrlline = NULL;
+	char *line = NULL, *tok, *savetok;
+	src = NIH_MUST( nih_sprintf(NULL, "%s/cgroup.subtree_control", from) );
+	dest = NIH_MUST( nih_sprintf(NULL, "%s/cgroup.subtree_control", to) );
+	line = read_oneline(src);
+	if (!line)
+		return true;
+
+	tok = strtok_r(line, " ", &savetok);
+	while (tok) {
+		NIH_MUST( nih_strcat_sprintf(&ctrlline, NULL, "+%s ", tok) );
+		tok = strtok_r(NULL, " ", &savetok);
+	}
+	free(line);
+	return write_string(dest, ctrlline);
+}
+
+bool unified_copy_controllers(const char *controller, const char *path)
+{
+	nih_local char *p = NULL;
+	char *pe;
+
+	if (!is_unified_controller(controller))
+		return true;
+
+	p = NIH_MUST( nih_strdup(NULL, path) );
+	pe = strrchr(p, '/');
+	if (pe)
+		*pe = '\0';
+	return do_copy_controllers(p, path);
+}
+
+bool create_leaf(const char *controller, const char *path, uid_t u, gid_t g)
+{
+	nih_local char *p = NULL;
+
+	if (!is_unified_controller(controller))
+		return true;
+
+	NIH_MUST( nih_strcat_sprintf(&p, NULL, "%s%s", path, U_LEAF) );
+	if (mkdir(p, 755) < 0 && errno != EEXIST)
+		return false;
+	if (mkdir(p, 755) < 0 && errno == EEXIST)
+		return true;
+	if (chown(p, u, g) < 0)
+		return false;
+	return true;
+}
+
+static bool copy_owner_perms_from_to(const char *from, const char *to)
+{
+	struct stat sb;
+	struct dirent dirent, *direntp;
+	bool error = false;
+	DIR *d;
+
+	if (stat(from, &sb) < 0)
+		return false;
+	if (chown(to, sb.st_uid, sb.st_gid) < 0)
+		return false;
+	if (chmod(to, sb.st_mode) < 0)
+		return false;
+
+	d = opendir(from);
+	if (!d)
+		return false;
+
+	while (readdir_r(d, &dirent, &direntp) == 0 && direntp) {
+		nih_local char  *srcp = NULL,
+			  *dstp = NULL;
+
+		if (!strcmp(direntp->d_name, ".") || !strcmp(direntp->d_name, ".."))
+			continue;
+		srcp = NIH_MUST( nih_sprintf(NULL, "%s/%s", from, direntp->d_name) );
+		dstp = NIH_MUST( nih_sprintf(NULL, "%s/%s", to, direntp->d_name) );
+		if (!file_exists(dstp))
+			continue;
+		if (stat(srcp, &sb) < 0)
+			continue;
+		if (chown(dstp, sb.st_uid, sb.st_gid) < 0) {
+			nih_error("Failed to chown file %s to %u:%u",
+					dstp, sb.st_uid, sb.st_gid);
+			error = true;
+		}
+		if (chmod(dstp, sb.st_mode) < 0) {
+			nih_error("Failed to chmod file %s to %o",
+					dstp, sb.st_mode);
+			error = true;
+		}
+	}
+	closedir(d);
+
+	return !error;
+}
+
+bool ensure_leafdir(const char *controller, const char *path)
+{
+	nih_local char *p = NIH_MUST( nih_sprintf(NULL, "%s%s", path, U_LEAF) );
+
+	if (mkdir(p, 0755) < 0) {
+		if (errno != EEXIST)
+			return false;
+		// existed, don't change perms
+		return true;
+	}
+	if (!copy_owner_perms_from_to(path, p)) {
+		rmdir(p);
+		return false;
+	}
+	return true;
+}
+
+void turn_mount_rw(const char *path)
+{
+	if (mount("", path, "cgroup", MS_REMOUNT, NULL) < 0) {
+		nih_error("Failed to turn %s mount readonly: %m", path);
+		exit(1);
+	}
+	nih_debug("Turned %s from ro->rw", path);
+}
+
+bool is_ro_mount(const char *path)
+{
+	struct statvfs sb;
+
+	if (statvfs(path, &sb) < 0) {
+		nih_error("statvfs failed on %s: %m", path);
+		exit(1);
+	}
+
+	return sb.f_flag & MS_RDONLY;
+}
